# General packages
library(tidyverse)
library(patchwork)

# Geo packages
library(terra)
library(sf)
library(geos)

# Stats packages
library(Rtsne)
library(cito)

source("R/utils.R")

load("data/r_objects/model_data.RData")
load("data/r_objects/ssdm_results.RData")
load("data/r_objects/msdm_embed_performance.RData")
load("data/r_objects/functional_groups.RData")

load("data/r_objects/sa_polygon.RData")
load("data/r_objects/range_maps.RData")

sf::sf_use_s2(use_s2 = FALSE)

model_data = model_data %>% 
  dplyr::filter(!is.na(fold_eval)) %>% 
  dplyr::mutate(species = as.factor(species))

# ------------------------------------------------------------------ #
# 1. Collect performance results                                  ####
# ------------------------------------------------------------------ #
performance = ssdm_results %>% 
  bind_rows(msdm_embed_performance) %>% 
  dplyr::group_by(species, model, metric) %>% 
  dplyr::summarise(value = mean(value, na.rm = F)) %>% 
  dplyr::mutate(
    metric = stringr::str_to_lower(metric),
    value = case_when(
      ((is.na(value) | is.nan(value)) & metric %in% c("auc", "f1", "accuracy", "precision", "recall")) ~ 0.5,
      ((is.na(value) | is.nan(value)) & metric %in% c("kappa")) ~ 0,
      .default = value
    )
  )

# ------------------------------------------------------------------ #
# 2. Model comparison                                             ####
# ------------------------------------------------------------------ #
## Overall model Comparison ####
df_plot = performance %>% 
  dplyr::filter(metric %in% c("auc", "f1", "kappa", "accuracy", "precision", "recall"))

ggplot(df_plot, aes(x = model, y = value, color = model)) +
  geom_boxplot(alpha = 0.5, outlier.shape = 16) + 
  facet_wrap(~ metric, scales = "free_y") +
  labs(
    title = "Model Performance Across Metrics",
    x = "Model",
    y = "Value",
    color = "Model",
  ) +
  theme_minimal(base_size = 14) +
  theme(
    strip.text = element_text(face = "bold"),
    axis.text.x = element_text(angle = 45, hjust = 1),
    legend.position = "right"
  )

ggsave("plots/publication/model_performance_across_metrics.pdf", 
       device = "pdf", 
       scale = 2,
       width = 18, 
       height = 14,
       units = "cm")

## Performance vs number of records ####
obs_count = model_data %>% 
  sf::st_drop_geometry() %>% 
  dplyr::filter(present == 1, !is.na(fold_eval)) %>% 
  dplyr::group_by(species) %>% 
  dplyr::summarise(obs = n())

df_plot = performance %>% 
  dplyr::filter(metric %in% c("auc", "f1", "kappa", "accuracy", "precision", "recall")) %>% 
  dplyr::left_join(obs_count, by = "species") 

ggplot(df_plot, aes(x = obs, y = value, color = model, fill = model)) +
  geom_point(alpha = 0.2) +
  geom_smooth() + 
  facet_wrap(~ metric, scales = "free_y") +
  scale_x_continuous(trans = "log10") +
  labs(
    title = "Model Performance vs Number of Records",
    x = "Number of Records (log scale)",
    y = "Value",
    color = "Model",
    fill = "Model"
  ) +
  theme_minimal(base_size = 14) +
  theme(
    strip.text = element_text(face = "bold"),
    axis.text.x = element_text(angle = 45, hjust = 1),
    legend.position = "right"
  )

ggsave("plots/publication/model_performance_vs_number_of_records.pdf", 
       device = "pdf", 
       scale = 2,
       width = 18, 
       height = 14,
       units = "cm")

## Performance vs functional group ####
load("data/r_objects/functional_groups.RData")

df_plot = performance %>% 
  dplyr::filter(metric %in% c("auc", "f1", "kappa")) %>% 
  dplyr::left_join(functional_groups, by = c("species" = "name_matched"))

ggplot(df_plot, aes(x = model, y = value, color = model)) +
  geom_boxplot(alpha = 0.5, outlier.shape = 16) + 
  facet_grid(rows = vars(metric), cols = vars(functional_group), scales = "free_y", switch = "both") +
  labs(
    title = "Model Performance Across Functional Groups",
    x = "Model",
    y = "Value",
    color = "Model",
  ) +
  theme_minimal(base_size = 14) +
  theme(
    strip.text = element_text(face = "bold"),
    panel.border = element_rect(color = "gray", fill = NA),
    axis.text.x = element_text(angle = 45, hjust = 1),
    legend.position = "right"
  )

ggsave("plots/publication/model_performance_vs_functional_groups.pdf", 
       device = "pdf", 
       scale = 2,
       width = 18, 
       height = 14,
       units = "cm")

# ------------------------------------------------------------------ #
# 3. Range predictions                                            ####
# ------------------------------------------------------------------ #
# Define plotting function
plot_predictions = function(spec, model_data, raster_data, algorithms = c("gam", "gbm", "rf", "nn", "msdm")){
  # Species data
  load("data/r_objects/range_maps.RData")
  
  pa_spec = model_data %>% 
    dplyr::filter(species == !!spec)
  range_spec = range_maps %>% 
    dplyr::filter(name_matched == !!spec) %>% 
    sf::st_transform(sf::st_crs(pa_spec))
  
  # Extract raster values into df
  bbox_spec = sf::st_bbox(pa_spec) %>% 
    expand_bbox(expansion = 0.25)
  
  raster_crop = terra::crop(raster_data, bbox_spec)
  new_data = raster_crop %>% 
    terra::values(dataframe = T, na.rm = T) %>% 
    dplyr::mutate(species = unique(pa_spec$species)) 
  
  # Make predictions
  for(algorithm in algorithms){
    message(algorithm)
    # Load model
    tryCatch({
      if(algorithm == "msdm"){
        load("data/r_objects/msdm_embed_results/msdm_embed_fit_full.RData")
        probabilities = predict(msdm_embed_fit, new_data, type = "response")[,1]
        predictions = factor(round(probabilities), levels = c("0", "1"), labels = c("A", "P"))
      } else if(algorithm == "nn") {
        load(paste0("data/r_objects/ssdm_results/models/", spec, "_nn_fit.RData"))
        probabilities = predict(nn_fit, new_data, type = "response")[,1]
        predictions = factor(round(probabilities), levels = c("0", "1"), labels = c("A", "P"))
      } else {
        load(paste0("data/r_objects/ssdm_results/models/", spec, "_", algorithm, "_fit.RData"))
        predictions = predict(get(paste0(algorithm, "_fit")), new_data, type = "raw")
      }
    }, error = function(e){
      warning(toupper(algorithm), ": Model could not be loaded.")
    })

    raster_pred = terra::rast(raster_crop, nlyrs = 1)
    values(raster_pred)[as.integer(rownames(new_data))] <- predictions
    
    # Plot
    plot(raster_pred, 
         type = "classes",
         levels = c("Absent", "Present"),
         main = paste0("Range prediction (", toupper(algorithm), "): ", spec))
    point_colors = sapply(pa_spec$present, function(x) ifelse(x == 0, "black", "white"))
    plot(pa_spec[,"present"], col = point_colors, add = T, pch = 16)
    plot(range_spec, border = "white", lwd = 2, col = NA, add = T)
  }
}

# Load raster
raster_filepaths = list.files("~/symobio-modeling/data/geospatial/raster/", full.names = T)
raster_data = terra::rast(raster_filepaths) %>% 
  terra::crop(sf::st_bbox(sa_polygon)) %>% 
  terra::project(sf::st_crs(model_data)$input)

spec = "Thyroptera tricolor"
pdf(file = paste0("plots/range_predictions/", spec, ".pdf"))
plot_predictions(spec, model_data, raster_data, algorithms = c("gam", "gbm", "rf", "msdm"))
dev.off()

# ------------------------------------------------------------------ #
# 5. Compare msdm predictions                                     ####
# ------------------------------------------------------------------ #
# Check predictions for different species
load("data/r_objects/msdm_embed_results/msdm_embed_fit.RData")
specs = sample(unique(model_data$species), size = 2, replace = F)

new_data = spatSample(raster_data, 1000, replace = F, as.df = T) %>% 
  drop_na()

new_data1 = dplyr::mutate(new_data, species = specs[1])
new_data2 = dplyr::mutate(new_data, species = specs[2])

pred1 = predict(msdm_embed_fit, new_data1, type = "response")
pred2 = predict(msdm_embed_fit, new_data2, type = "response")

pred1 == pred2

# ---> identical predictions

# Permute embeddings
embeddings = coef(msdm_embed_fit)[[1]][[1]]
msdm_embed_fit$net$e_1$parameters$weight$set_data(torch::torch_tensor(embeddings[sample.int(nrow(embeddings)), ], device = msdm_embed_fit$net$e_1$parameters$weight$device, dtype = msdm_embed_fit$net$e_1$parameters$weight$dtype ))

# ---> Invalid external pointer

# ------------------------------------------------------------------ #
# 6. Compare species embeddings                                   ####
# ------------------------------------------------------------------ #
all_embeddings = lapply(1:5, function(fold){
  load(paste0("data/r_objects/msdm_embed_results/msdm_embed_fit_fold", fold, ".RData"))
  coef(msdm_embed_fit)[[1]][[1]]
})

pairs = combn(1:5, 2, simplify = F)
pairwise_rv = lapply(pairs, function(pair){
  FactoMineR::coeffRV(all_embeddings[[pair[1]]], all_embeddings[[pair[2]]])  
})

rv_matrix = matrix(NA, 5, 5)
for(i in seq_along(pairs)){
  rv_matrix[pairs[[i]][2], pairs[[i]][1]] = pairwise_rv[[i]][["rv"]]
}

p_matrix = matrix(NA, 5, 5)
for(i in seq_along(pairs)){
  p_matrix[pairs[[i]][2], pairs[[i]][1]] = pairwise_rv[[i]][["p.value"]]
}

# ------------------------------------------------------------------ #
# 7. Analyse species embeddings                                      ####
# ------------------------------------------------------------------ #
embeddings = coef(msdm_fit_embedding_raw)[[1]][[1]]
species_lookup = msdm_fit_embedding_raw$data$data[,c("species", "species_int")] %>% 
  distinct() %>% 
  mutate(genus = stringr::str_extract(rownames(embeddings), "([a-zA-Z]+) (.*)", group = 1))

rownames(embeddings) = species_lookup$species

##  Dimensionality reduction        ####
### PCA                             ####
pca_result = prcomp(t(embeddings), scale. = TRUE)
var_explained = pca_result$sdev^2 / sum(pca_result$sdev^2)
plot(var_explained)

# --> Variance explained is distributed rather evenly across dimensions
# --> Dimensionality reduction probably not useful

coords = pca_result$rotation[,1:2] %>%
  magrittr::set_colnames(c("X", "Y"))

df_plot = species_lookup %>% 
  left_join(functional_groups, by = c("species" = "name_matched")) %>% 
  bind_cols(coords)

ggplot(df_plot, aes(x = X, y = Y, col = functional_group, label=genus)) +
  geom_point() +
  geom_text(hjust=0, vjust=0) +
  guides(col="none")

# --> PCA not informative on 2 dimensions either

### TSNE                            ####
tsne_result <- Rtsne(embeddings, verbose = FALSE)
coords = tsne_result$Y %>%
  magrittr::set_colnames(c("X", "Y"))

df_plot = species_lookup %>% 
  left_join(functional_groups, by = c("species" = "name_matched")) %>% 
  bind_cols(coords)

ggplot(df_plot, aes(x = X, y = Y, col = functional_group, label=genus)) +
  geom_point() +
  geom_text(hjust=0, vjust=0) +
  guides(col="none")

# --> TSNE not informative on 2 dimensions either


### KNN                             ####
embeddings_dist = as.matrix(dist(embeddings))

k = 9
knn_results = sapply(as.character(species_lookup$species), FUN = function(spec){
  return(sort(embeddings_dist[spec,])[2:(k+1)])
}, USE.NAMES = T, simplify = F)

plot_knn_ranges = function(spec){
  spec_df = model_data %>% 
    dplyr::filter(species == !!spec)
  
  p_spec_range = ggplot() +
    geom_sf(data = st_as_sf(sa_polygon)) +
    geom_sf(data = spec_df) + 
    ggtitle(paste0(spec)) +
    theme_minimal()
  
  knn_df = model_data %>% 
    dplyr::filter(species %in% c(names(knn_results[[spec]])))
  
  p_knn_ranges = ggplot() +
    geom_sf(data = st_as_sf(sa_polygon)) +
    geom_sf(data = knn_df, aes(color = species)) + 
    facet_wrap(facets = "species", ncol = 3) +
    theme_minimal() +
    guides(color="none")
  
  p_spec_range + p_knn_ranges
}

plot_knn_ranges(sample(species_lookup$species, 1)) # Repeat this line to plot random species

# Clustering
embeddings_dist = dist(embeddings, method = "euclidean")
hclust_complete = hclust(dist_matrix, method = "complete")
plot(hclust_complete, main = "Complete Linkage", xlab = "", sub = "", labels = FALSE)

# Phylo Correlation
load("data/r_objects/phylo_dist.RData")

species_intersect = intersect(colnames(phylo_dist), species_lookup$species)

phylo_dist_subset = phylo_dist[species_intersect, species_intersect]

e_indices = match(species_intersect, species_lookup$species)
embeddings_dist_subset = as.matrix(embeddings_dist)[e_indices, e_indices]
FactoMineR::coeffRV(embeddings_dist_subset, phylo_dist_subset)

# Trait Correlation
load("data/r_objects/func_dist.RData")

species_intersect = intersect(colnames(func_dist), species_lookup$species)

func_dist_subset = func_dist[species_intersect, species_intersect]

e_indices = match(species_intersect, species_lookup$species)
embeddings_dist_subset = as.matrix(embeddings_dist)[e_indices, e_indices]
FactoMineR::coeffRV(embeddings_dist_subset, func_dist_subset)

# Range Correlation
load("data/r_objects/range_dist.RData")

species_intersect = intersect(colnames(range_dist), species_lookup$species)

range_dist_subset = range_dist[species_intersect, species_intersect]

e_indices = match(species_intersect, species_lookup$species)
embeddings_dist_subset = as.matrix(embeddings_dist)[e_indices, e_indices]
FactoMineR::coeffRV(embeddings_dist_subset, range_dist_subset)

