### Range coverage bias

Range coverage bias was calculated as 1 minus the ratio of the actual range coverage and the hypothetical range coverage if all observations were maximally spread out across the range.

$$
  RangeCoverageBias = 1 - \frac{RangeCoverage}{min({N_{obs\_total}} / {N_{cells\_total}}, 1)}
  $$
    
    Higher bias values indicate that occurrence records are spatially more clustered within the range of the species.
  
  -   There was no strong relationship between range coverage bias and model performance
  
  ```{r range_coverage_bias, echo = FALSE, message=FALSE, warnings=FALSE, eval=F}
  df_occs_total = occs_final %>% 
    st_drop_geometry() %>% 
    group_by(species) %>% 
    summarise(occs_total = n())
  
  df_join = df_occs_total %>% 
    dplyr::inner_join(df_cells_total, by = "species") %>% 
    dplyr::inner_join(df_cells_occ, by = "species") %>% 
    dplyr::mutate(range_bias = 1-((cells_occupied / cells_total) / pmin(occs_total / cells_total, 1)))
  
  df_plot = performance %>% 
    inner_join(df_join, by = "species")
  
  # Calculate regression lines for each model and metric combination
  suppressWarnings({
    regression_lines = df_plot %>%
      group_by(model, metric) %>%
      group_modify(~lin_regression(.x$range_bias, .x$value))
  })
  
  # Create base plot
  plot <- plot_ly() %>% 
    layout(
      title = "Model Performance vs. Range coverage bias",
      xaxis = list(title = "Range coverage bias"),
      yaxis = list(title = "Value"),
      legend = list(x = 1.1, y = 0.5),  # Move legend to the right of the plot
      margin = list(r = 150),  # Add right margin to accommodate legend
      hovermode = 'closest',
      updatemenus = list(
        list(
          type = "dropdown",
          active = 0,
          buttons = plotly_buttons
        )
      )
    )
  
  # Add regression lines and confidence intervals for each model
  for (model_name in unique(df_plot$model)) {
    plot <- plot %>%
      add_markers(
        data = filter(df_plot, model == model_name),
        x = ~ range_bias,
        y = ~ value,
        color = model_name,  # Set color to match legendgroup
        legendgroup = model_name,
        opacity = 0.6,
        name = ~model,
        hoverinfo = 'text',
        text = ~paste("Species:", species, "<br>Range coverage bias:", range_bias, "<br>Value:", round(value, 3)),
        transforms = list(
          list(
            type = 'filter',
            target = ~metric,
            operation = '=',
            value = focal_metrics[1]
          )
        )
      )
  }
  
  for (model_name in unique(df_plot$model)) {
    reg_data = dplyr::filter(regression_lines, model == model_name)
    plot = plot %>% 
      add_lines(
        data = reg_data,
        x = ~x,
        y = ~fit,
        color = model_name,  # Set color to match legendgroup
        legendgroup = model_name,
        name = paste(model_name, '(fit)'),
        showlegend = FALSE,
        transforms = list(
          list(
            type = 'filter',
            target = ~metric,
            operation = '=',
            value = focal_metrics[1]
          )
        )
      )
  }
  
  
  bslib::card(plot, full_screen = T)
  ```
  
  ### Functional group
  
  Functional groups were assigned based on taxonomic order. The following groupings were used:
    
    | Functional group      | Taxomic orders                                                        |
    |-------------------|-----------------------------------------------------|
    | large ground-dwelling | Carnivora, Artiodactyla, Cingulata, Perissodactyla                    |
    | small ground-dwelling | Rodentia, Didelphimorphia, Soricomorpha, Paucituberculata, Lagomorpha |
    | arboreal              | Primates, Pilosa                                                      |
    | flying                | Chiroptera                                                            |
    
    -   Models for bats tended to perform slightly worse than for other groups.
  
  ```{r functional_groups, echo = FALSE, message=FALSE, warnings=FALSE, eval=F}
  df_plot = performance %>% 
    dplyr::left_join(functional_groups, by = c("species" = "name_matched"))
  
  plot <- plot_ly(
    data = df_plot,
    x = ~functional_group,
    y = ~value,
    color = ~model,
    type = 'box',
    boxpoints = "all",
    jitter = 1,
    pointpos = 0,
    hoverinfo = 'text',
    text = ~paste("Species:", species, "<br>Functional group:", functional_group, "<br>Value:", round(value, 3)),
    transforms = list(
      list(
        type = 'filter',
        target = ~metric,
        operation = '=',
        value = focal_metrics[1]  # default value
      )
    )
  )
  
  plot <- plot %>%
    layout(
      title = "Model Performance vs. Functional Group",
      xaxis = list(title = "Functional group"),
      yaxis = list(title = "Value"),
      legend = list(x = 1.1, y = 0.5),  # Move legend to the right of the plot
      margin = list(r = 150),  # Add right margin to accommodate legend
      hovermode = 'closest',
      boxmode = "group",
      updatemenus = list(
        list(
          type = "dropdown",
          active = 0,
          buttons = plotly_buttons
        )
      )
    )
  
  bslib::card(plot, full_screen = T)
  ```
  